/*
 * Power Switch driver for EyeBot M6.
 *
 * Copyright (C) 2006 Bernard Blackham <bernard@blackham.com.au>
 *
 * Licensed under the GPLv2.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>
#include <linux/sysctl.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/reboot.h>
#include <linux/pm.h>

#include <asm/uaccess.h>

#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>




/****** User tweakable variables. ******/

/* Which GPIO pin is the power switch on? */
#define PS_GPIO 80

/* Which GPIO do we set low to turn power off? */
#define PS_KILL_GPIO 32

/* How long do we wait for a debounce? */
#define PS_DEBOUNCE_JIFFIES (HZ/10) /* 100ms */

/* How many events can queue up in userspace before we start dropping them? */
#define PS_QUEUE_LEN    8

/* How long does the power button need to be held down for to force power off? */
#define FORCE_POWEROFF_SECONDS 1
#define FORCE_POWEROFF_JIFFIES (FORCE_POWEROFF_SECONDS * HZ)

/**** End of user tweakable variables. No user serviceable parts below.  ****/



/****** Macros ******/

#define DEVICE_NAME "powerswitch"
#define MODULE_NAME "powerswitch"
#define VERSION_STR "0.10"

#if 0
#define DEBUG(s, x...) printk(MODULE_NAME ": " s "\n", ##x)
#else
#define DEBUG(x...) do { } while (0)
#endif

/****** Definitions ******/

typedef int ps_event_t;
typedef struct tag_ps_queue {
	ps_event_t pending_msgs[PS_QUEUE_LEN];
	int read_index;
	int write_index;
	wait_queue_head_t wait_queue;
	spinlock_t lock;

	struct tag_ps_queue *next;
} ps_queue_t;

/****** Statics ******/

static char init_banner[] __initdata =
	KERN_INFO "M6 Power Switch Driver "VERSION_STR" loaded\n";
static char exit_banner[] __exitdata =
	KERN_INFO "M6 Power Switch Driver "VERSION_STR" unloaded\n";

static ps_queue_t *ps_queue = NULL;
static spinlock_t ps_queue_lock = SPIN_LOCK_UNLOCKED;

static struct class *ps_class = NULL;
static struct cdev ps_cdev;
static dev_t ps_devnum = 0;

static void (*pm_power_off_save)(void) = NULL;

static void ps_forceoff(unsigned long data);
static void ps_poweroff(unsigned long data);
static DEFINE_TIMER(ps_poweroff_timer, ps_forceoff, 0, 0);

static void ps_pushed(unsigned long data);
static DEFINE_TIMER(ps_down_timer, ps_pushed, 0, 1);
static DEFINE_TIMER(ps_up_timer, ps_pushed, 0, 0);

static int awaiting_poweroff = 0;
static uint64_t ps_last_pressed = 0; /* if down, jiffies of first down event. */

/****** Functions ******/

/*
 * eyebot_power_off
 *
 * The kernel calls this function when it wants to power down the Eyebot.
 * Toggles the appropriate GPIO lines so that we lose power.
 */
static void eyebot_power_off(void)
{
	printk(KERN_EMERG "Eyebot powering off!\n");

	/* Lower the pin that keeps us alive... */
	GPCR(PS_KILL_GPIO) = GPIO_bit(PS_KILL_GPIO);

	for(;;)
		cpu_relax();
}

/*
 * add_ps_watcher
 *
 * Adds a new watcher to be notified of changes in power button state.
 *
 * Returns a handle for the watcher.
 */
static ps_queue_t *add_ps_watcher(void)
{
	unsigned long flags;

	ps_queue_t *q = kmalloc(sizeof(*q), GFP_KERNEL);
	if (!q)
		return NULL;

	memset(q, 0, sizeof(*q));

	init_waitqueue_head(&q->wait_queue);
	spin_lock_init(&q->lock);

	spin_lock_irqsave(&ps_queue_lock, flags);
	q->next = ps_queue;
	ps_queue = q;
	spin_unlock_irqrestore(&ps_queue_lock, flags);

	return q;
}

/*
 * remove_ps_watcher
 *
 * Given a handle, removes this watcher from the list.
 */
static void remove_ps_watcher(ps_queue_t *q)
{
	unsigned long flags;
	int found = 0;
	ps_queue_t *cur, *prev = NULL;

	spin_lock_irqsave(&ps_queue_lock, flags);

	for (cur = ps_queue; cur; cur = cur->next) {
		if (cur == q) {
			if (prev != NULL)
				prev->next = cur->next;
			else
				ps_queue = cur->next;
			kfree(cur);
			found = 1;
			break;
		}
		prev = cur;
	}

	spin_unlock_irqrestore(&ps_queue_lock, flags);

	BUG_ON(found == 0);
}

/*
 * ps_queue_one_event
 *
 * Enqueues an event into the given watcher's queue.
 */
static void ps_queue_one_event(ps_queue_t *q, const ps_event_t *event)
{
	unsigned long flags;

	spin_lock_irqsave(&q->lock, flags);

	if (q->read_index == ((q->write_index+1) % PS_QUEUE_LEN)) {
		printk(KERN_WARNING MODULE_NAME ": queue overflowed.\n");
	} else {
		q->pending_msgs[q->write_index++] = *event;
		if (q->write_index >= PS_QUEUE_LEN)
			q->write_index = 0;
	}

	spin_unlock_irqrestore(&q->lock, flags);
}

/*
 * ps_dequeue_one_event
 *
 * Returns an event for the given watcher's queue.
 */
static int ps_dequeue_one_event(ps_queue_t *q, ps_event_t *event)
{
	unsigned long flags;
	int found = 0;

	spin_lock_irqsave(&q->lock, flags);

	if (q->read_index != q->write_index) {
		*event = q->pending_msgs[q->read_index++];

		if (q->read_index >= PS_QUEUE_LEN)
			q->read_index = 0;

		found = 1;
	}

	spin_unlock_irqrestore(&q->lock, flags);

	return found;
}

/*
 * ps_queue_event
 *
 * Enqueue an event into queues of all watchers.
 */
static void ps_queue_event(const ps_event_t *event)
{
	ps_queue_t *cur;
	unsigned long flags;

	if (ps_queue == NULL)
	{
		//printk(KERN_INFO MODULE_NAME ": unwanted event 0x%x\n", *event);
		return;
	}

	spin_lock_irqsave(&ps_queue_lock, flags);

	for (cur = ps_queue; cur; cur = cur->next) {
		ps_queue_one_event(cur, event);
		wake_up_interruptible(&cur->wait_queue);
	}

	spin_unlock_irqrestore(&ps_queue_lock, flags);
}

/*
 * ps_pushed
 *
 * Called when a debounce timer expires on a button push or release.
 * Notifies all watchers of the relevant event.
 */
static void ps_pushed(unsigned long down)
{
	ps_event_t event;

	/* Ignore spurious state changes. */
	if (ps_last_pressed == 0 && !down)
		return;
	if (ps_last_pressed != 0 && down)
		return;

	if (awaiting_poweroff && !down)
		ps_poweroff(0);

	if (down) {
		ps_last_pressed = jiffies;
		event = 0;
	} else {
		/* Scale down to userspace jiffies. */
		event = (int)((jiffies - ps_last_pressed) * (100 / HZ));
		ps_last_pressed = 0;
	}

	ps_queue_event(&event);
}

/*
 * ps_forceoff
 *
 * Called when the force powerdown timer expires. ie, the power button has
 * been held for N seconds and we just want to cut the power. We wait for the
 * power button to be released before powering off.
 */
static void ps_forceoff(unsigned long data)
{
	printk(KERN_EMERG "Power-off forced.\n");
	awaiting_poweroff = 1;
}

/*
 * ps_poweroff
 *
 * Causes an emergency power off.
 */
static void ps_poweroff(unsigned long data)
{
	kernel_power_off();
	for(;;)
		cpu_relax();
}

/*
 * ps_interrupt
 *
 * Our interrupt routine when the button is pushed or released.
 */
static irqreturn_t ps_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int down;

	/* Figure out if the switch is up or down.
	 * FIXME: is there a better way that doesn't race against the IRQ handler?
	 */
	down = (GPLR(PS_GPIO) & GPIO_bit(PS_GPIO));

	/* Get the force power-off timer ready. */
	if (down && !timer_pending(&ps_poweroff_timer)) {
		/* Start the countdown for doom. */
		ps_poweroff_timer.expires = jiffies + FORCE_POWEROFF_JIFFIES;
		add_timer(&ps_poweroff_timer);
	} else if (!down && timer_pending(&ps_poweroff_timer)) {
		/* Abort! Abort! Abort! */
		del_timer(&ps_poweroff_timer);
	}

	/* Re-arm the debounce timers. */
	if (down) {
		if (timer_pending(&ps_down_timer))
			del_timer(&ps_down_timer);
		if (timer_pending(&ps_up_timer))
			del_timer(&ps_up_timer);

		ps_down_timer.expires = jiffies + PS_DEBOUNCE_JIFFIES;

		add_timer(&ps_down_timer);
	} else {
		if (timer_pending(&ps_up_timer))
			del_timer(&ps_up_timer);

		ps_up_timer.expires = jiffies + PS_DEBOUNCE_JIFFIES;

		add_timer(&ps_up_timer);
	}

	return IRQ_HANDLED;
}

/*
 * ps_dummy
 *
 * Dummy interrupt function for the kill line.
 */
static irqreturn_t ps_dummy(int irq, void *dev_id, struct pt_regs *regs)
{
	return IRQ_HANDLED;
}

/*
 * ps_open
 *
 * Called when a user opens our device node.
 */
static int ps_open(struct inode *inode, struct file *filp)
{
	/* Register for notifications on the power switch. */
	if ((filp->private_data = (void*)add_ps_watcher()) == NULL)
		return -ENOMEM;

	return 0;
}

/*
 * ps_read
 *
 * Performs a read from the device node. If blocking, waits for an event and
 * returns a single event to userspace. If non-blocking, returns an event if
 * one exists, or nothing at all.
 */
static ssize_t ps_read(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	int ret;
	ps_event_t event;
	ps_queue_t *q;

	if (count == 0)
		return 0;

	q = (ps_queue_t*)filp->private_data;

	if ((filp->f_flags & O_NONBLOCK) &&
			q->read_index == q->write_index)
		return -EAGAIN;

	ret = wait_event_interruptible(q->wait_queue,
			(q->read_index != q->write_index));
				
	if (ret != 0)
		return ret;

	ps_dequeue_one_event(q, &event);

	if (count >= sizeof(event)) {
		if (copy_to_user(buf, &event, sizeof(event)) != 0)
			return -EFAULT;
		return sizeof(event);
	}
	else
		return 0;
}

/*
 * ps_release
 *
 * Called when the device node is closed.
 */
static int ps_release(struct inode *inode, struct file *filp)
{
	/* Unregister for notifications on the power switch. */
	remove_ps_watcher((ps_queue_t*)filp->private_data);
	return 0;
}

static struct file_operations ps_fops =
{
	owner:   THIS_MODULE,

	open:    ps_open,
	release: ps_release,
	read:    ps_read,
};

/*
 * powerswitch_init
 *
 * Register ourselves with the relevant subsystems.
 */
static int __init powerswitch_init(void)
{
	int ret;

	/* Capture our power line to ensure it can't be probed and disabled accidentally. */
	ret = request_irq(IRQ_GPIO(PS_KILL_GPIO), ps_dummy,
			SA_INTERRUPT, DEVICE_NAME ":lifeline", NULL);
	if (ret != 0) {
		printk(KERN_WARNING MODULE_NAME ": Unable to request IRQ %d, err: %d\n", PS_KILL_GPIO, ret);
		return ret;
	}

	pxa_gpio_mode(PS_GPIO | GPIO_IN);
	set_irq_type(IRQ_GPIO(PS_GPIO), IRQT_BOTHEDGE);

	ret = request_irq(IRQ_GPIO(PS_GPIO), ps_interrupt,
			SA_INTERRUPT, DEVICE_NAME":button", NULL);
	if (ret != 0) {
		printk(KERN_WARNING MODULE_NAME ": Unable to request IRQ %d, err: %d\n", PS_GPIO, ret);
		return ret;
	}

	if ((ret = alloc_chrdev_region(&ps_devnum, 0, 1, DEVICE_NAME)) < 0) {
		printk(KERN_WARNING MODULE_NAME ": Unable to allocate major, err: %d\n", ret);
		return ret;
	}

	cdev_init(&ps_cdev, &ps_fops);
	ps_cdev.owner = THIS_MODULE;

	if ((ret = cdev_add(&ps_cdev, ps_devnum, 1)) != 0)
	{
		printk(KERN_WARNING MODULE_NAME ": cdev_add failed: %d\n", ret);
		return ret;
	}

	ps_class = class_create(THIS_MODULE, DEVICE_NAME);
	if (IS_ERR(ps_class))
	{
		printk(KERN_WARNING MODULE_NAME ": Unable to create class\n");
		return -1;
	}

	class_device_create(ps_class, NULL, ps_devnum, NULL, DEVICE_NAME);

	pm_power_off_save = pm_power_off;
	pm_power_off = eyebot_power_off;

	printk(init_banner);

	return 0;
}

/*
 * powerswitch_exit
 *
 * Unregister ourselves with the relevant subsystems. It's guaranteed that we
 * won't be called with users on the device nodes still, so we know all
 * watchers have been cleaned up.
 */
static void __exit powerswitch_exit(void)
{
	pm_power_off = pm_power_off_save;

	class_device_destroy(ps_class, ps_devnum);
	class_destroy(ps_class);

	cdev_del(&ps_cdev);

	unregister_chrdev_region(ps_devnum, 1);

	free_irq(IRQ_GPIO(PS_KILL_GPIO), NULL);
	free_irq(IRQ_GPIO(PS_GPIO), NULL);

	printk(exit_banner);
}

module_init(powerswitch_init);
module_exit(powerswitch_exit);

MODULE_AUTHOR("Bernard Blackham");
MODULE_DESCRIPTION("Power switch driver for the Eyebot M6");
MODULE_LICENSE("GPL");
MODULE_VERSION(VERSION_STR);
